from pyproj import CRS
import logging
import geopandas as gpd
def is_wkt_string(value: str) -> bool:
    """
    Check if the provided string is a valid WKT representation of a CRS.

    Args:
    - value (str): The string to be tested.

    Returns:
    - bool: True if the string is a valid WKT, False otherwise.
    """
    try:
        # Try to create a CRS object from the WKT string
        crs = CRS.from_wkt(value)
        logging.info("Valid WKT string for CRS.")
        return True
    except Exception as e:
        logging.error(f"Invalid WKT string: {e}")
        return False
def get_crs(crs):
    if crs is not None and isinstance(crs,str) and os.path.isfile(crs):
        crs = get_df(crs)
    if isinstance(crs, gpd.GeoDataFrame):
        try:
            crs = crs.crs
        except Exception as e:
            logging.warning(f"crs conversion failed: {e}")
    if is_wkt_string(crs):
        crs = CRS.from_wkt(crs)
    return crs
    
def get_wkt_string(crs):
    # Extract the CRS if provided as a GeoDataFrame
    if isinstance(crs, CRS):
        wkt_string = crs.to_wkt()
    else:
        # Assume the input is a WKT string and attempt to parse it
        wkt_string = CRS.from_wkt(crs).to_wkt()
    return wkt_string
def correct_prj(prj_path, default_epsg=4326):
    is_valid, crs = validate_prj(prj_path)
    if is_valid:
        return crs

    logging.warning(f"Attempting to set default EPSG code: {default_epsg}")
    corrected_crs = CRS.from_epsg(default_epsg)

    # Save the corrected CRS
    with open(prj_path, "w") as file:
        file.write(corrected_crs.to_wkt())

    return corrected_crs

def validate_prj(prj_path):
    try:
        logging.warning(f"validating .prj: {prj_path}")
        # Attempt to load the CRS
        with open(prj_path, "r") as file:
            prj_content = file.read()
        crs = CRS.from_wkt(prj_content)
        logging.info("CRS is valid.")
        return True, crs
    except Exception as e:
        logging.error(f"Invalid CRS: {e}")
        return False, None
def save_valid_crs(crs: CRS, file_path: str):
    """
    Save the CRS to a .prj file if it is valid.

    Args:
        crs: A valid pyproj CRS object, a GeoDataFrame, or WKT string.
        file_path: Path to save the .prj file.
    """
    try:
        crs = get_crs(crs)
        wkt_string = get_wkt_string(crs)
        # Write the WKT string to the specified file
        with open(file_path, "w") as file:
            file.write(wkt_string)

        # Validate by reading it back
        loaded_crs = CRS.from_wkt(wkt_string)
        logging.info(f"Successfully saved and verified CRS at {file_path}")

    except Exception as e:
        logging.error(f"Error saving CRS: {e}")




from .excel_module import get_df,get_row,get_cell_value,gpd,eatAll
from .file_utils import safe_excel_save
from shapely.geometry import Polygon,MultiPolygon
import os
class SingletonMeta(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(SingletonMeta, cls).__call__(*args, **kwargs)
        return cls._instances[cls]
def get_project_name(filePath):
    directory = os.path.dirname(filePath)
    baseName = os.path.basename(filePath)
    fileName,ext = os.path.splitext(baseName)
    return directory,fileName,ext
def create_file_name(filePath,suffix):
    suffix = eatAll(suffix,'.')
    directory,fileName,ext=get_project_name(filePath)
    return os.path.join(directory,f"{fileName}.{suffix}")
def update_file(df,compFilePath,suffix):
    suffix = eatAll(suffix,'.')
    filePath = create_file_name(compFilePath,suffix)
    safe_excel_save(df,filePath)
    return filePath
class landManager(metaclass=SingletonMeta):
    def __init__(self,directory_js=None,default_polygon=None,column=None,value=None,epsg=4326,index=1):
        if not hasattr(self, 'initialized'):
            self.initialized = True
            directory_js = directory_js or {}
            self.file_manager={}
            self.closest_points={}
            self.default_polygon= default_polygon
            for directory_name,directory_path in directory_js.items():
                self.file_manager[directory_name] = {} 
                self.ext_js = {"shp":'.shp', "prj":'.prj', "cpg":'.cpg', "dbf":'.dbf', "shx":'.shx','geojson':'.geojson'}
                for file_item in os.listdir(directory_path):
                    item_path = os.path.join(directory_path,file_item)
                    self.file_manager[directory_name][os.path.splitext(file_item)[-1][1:]] = item_path
                if self.default_polygon == None:
                    gdf=self.get_contents(directory_name,file_type='shp')
                    if isinstance(gdf,gpd.GeoDataFrame):
                        self.default_polygon = self.get_polygon(gdf,column=column,value=value,index=index)
                self.check_geo_json(directory_name,epsg=epsg)
    @staticmethod
    def get_instance(directory_js=None):
        if landManager._instance is None:
            landManager(directory_js)
        return landManager._instance
    def get_polygon(self,gdf,column=None,value=None,index=1):
        if isinstance(gdf,Polygon) or isinstance(gdf,MultiPolygon):
            return gdf
        gdf = get_df(gdf)
        polygon = get_cell_value(gdf,'geometry',index)
        if polygon:
            return polygon
        if column or value:
            gdf,index = get_row(target_value=value,column_name=column,df=gdf)
        if isinstance(gdf,gpd.GeoDataFrame):
            return get_cell_value(gdf,'geometry',index)
    def check_geo_json(self,dir_name,epsg=4326):
        geoJsonPath = self.file_manager.get(dir_name,{}).get('geojson')
        if geoJsonPath == None or not os.path.isfile(geoJsonPath):
            gdf = self.get_contents(dir_name,'shp')
            geoJsonPath = self.update_file(gdf,dir_name,'geojson')
        return geoJsonPath
    def update_file(self,df,dir_name,suffix):
        file_path = self.get_file_path(dir_name,'shp')
        if file_path and os.path.isfile(file_path):
            filePath = update_file(df,file_path,suffix)
            if filePath:
                self.file_manager[dir_name][suffix] = filePath
            return filePath
        return file_path
    def get_file_path(self,dir_name,file_type):
        if dir_name == None:
            dir_name = list(self.file_manager.keys())
            if dir_name:
                dir_name = dir_name[0]
            else:
                return None
        file_path = self.file_manager.get(dir_name,{}).get(file_type)
        if file_path and os.path.isfile(file_path):
            return file_path
    def get_contents(self,dir_name,file_type,update=None):
        if update is not None and not isinstance(update,bool):
            file_path = self.update_file(update,dir_name,file_type)
        if os.path.isfile(file_type):
            file_path = file_type
        else:
            file_path = self.get_file_path(dir_name,file_type)
        if file_path:
            
            try:
                data = get_df(file_path)
                return data
            except Exception as e:
                print(f"{e}")
def get_proj(dir_name=None,file_path="prj",update=False):
    land_mgr = landManager.get_instance()
    return land_mgr.get_contents(dir_name,file_path,update=update)
def get_dbf(dir_name=None,file_path="dbf",update=False):
    land_mgr = landManager.get_instance()
    return land_mgr.get_contents(dir_name,file_path,update=update)
def get_shp(dir_name=None,file_path="shp",update=False):
    land_mgr = landManager.get_instance()
    return land_mgr.get_contents(dir_name,file_path,update=update)
def get_cpg(dir_name=None,file_path="cpg",update=False):
    land_mgr = landManager.get_instance()
    return land_mgr.get_contents(dir_name,file_path,update=update)
def get_dbf(dir_name=None,file_path="dbf",update=False):
    land_mgr = landManager.get_instance()
    return land_mgr.get_contents(dir_name,file_path,update=update)
def get_shx(dir_name=None,file_path="shx",update=False):
    land_mgr = landManager.get_instance()
    return land_mgr.get_contents(dir_name,file_path,update=update)
def get_geojson(dir_name=None,file_path="geojson",update=False):
    land_mgr = landManager.get_instance()
    return land_mgr.get_contents(dir_name,file_path,update=update)
def get_bp(bp,column=None,value=None,index=1):
    if isinstance(bp,Polygon) or isinstance(bp,MultiPolygon):
        return bp
    land_mgr = landManager.get_instance()
    gdf = get_shp(dir_name=dir_name,file_path="shp")
    bp = land_mgr.get_polygon(gdf,column=column,value=value,index=index) 
    return bp



from .file_utils import *
def is_inverse(series,inverse=False):
    if inverse:
        series = ~series
    return series
def update_excel(df,header,query):
    df = get_df(df)
    query = make_type(query,[str])
    df = convert_column(df,header,[str])
    return df,query
def merge_dataframes(dataframes):
    merged_df = pd.concat(dataframes, ignore_index=True)
    return merged_df
def excel_to_dict(df):
    # Read the Excel file
    df = get_df(df)
    # Convert each row to a dictionary with column headers as keys
    rows_as_dicts = df.to_dict(orient='records')
    return rows_as_dicts
def get_headers(df):
    df = get_df(df)
    column_names = df.columns.tolist()
    return column_names
def get_row_as_list(df,index=0):
    df=get_df(df)
    if get_row_number(df)>index:
        return df.loc[index].astype(str).tolist()
def get_row_number(df):
    df=get_df(df)
    return len(df)
def get_int_from_column(obj,column):
    headers = get_headers(obj)
    return get_itter(headers,column)
def get_column_from_int(obj,i):
    headers = get_headers(obj)
    column = safe_itter_get(headers,i)
    # Return the column headers
    return column
def find_row_with_matching_cell(excel_datas,search_column='',search_value=''):
    matching_row = [excel_data for excel_data in excel_datas if isinstance(excel_data,dict) and excel_data.get(search_column) == search_value]
    if matching_row and isinstance(matching_row,list) and len(matching_row)>0:
        return matching_row[0]
    return {}
def get_itter(list_obj,target):
    for i,string in enumerate(list_obj):
        if string == target:
            return i
    return None
def safe_itter_get(list_obj,i):
    if len(list_obj)>i:
        return list_obj[i]
def get_expected_headers(df,*expected_headers):
    df = get_df(df)
    if isinstance(expected_headers,tuple or set):
        expected_headers=list(expected_headers)
    else:
        expected_headers=make_list(expected_headers)
    expected_headers = {expected_header:"" for expected_header in expected_headers}
    return get_closest_headers(df,expected_headers)
def get_closest_headers(df,expected_headers={}):
    actual_headers = get_headers(df)  # Extract the actual headers from the DataFrame
    # Mapping actual headers to expected headers based on closest match
    for expected_header in expected_headers:
        # Using get_close_matches to find the closest match; returns a list
        close_matches = get_close_matches(expected_header, actual_headers, n=1, cutoff=0.6)
        if close_matches:
            expected_headers[expected_header] = close_matches[0]
        else:
            # If no close matches found, leave as empty string which signifies no match found
            expected_headers[expected_header] = ""
    return expected_headers
def find_closest_header(df,header):
    df = get_df(df)
    df_headers= df.columns.tolist()
    header_lower = header.lower()
    for df_header in df_headers:
        df_header_lower = df_header.lower()
        if df_header_lower in header_lower or header_lower in df_header_lower:
            js={header_lower:df_header}
            return js
def get_first_for_each(df,headers,queries,new_file_path=None):
    df=get_df(df)
    headers = get_expected_headers(df,headers).values()
    # Filter the DataFrame to only include rows with ZIP codes that are in the 'zips' list
    df=filter_and_deduplicate_df(df, headers, queries, dedup_columns=None)
    # Save the filtered and deduplicated DataFrame to a new Excel file
    return df
def get_cell_value(df, column_header, row_index=1):
    """
    Retrieves the value from a specified cell in the GeoDataFrame.
    
    :param df: GeoDataFrame or filepath to the shapefile.
    :param column_header: The header of the column from which to retrieve the value.
    :param row_index: The index of the row from which to retrieve the value.
    :return: The value located at the specified column and row.
    """
    df = get_df(df)
    # Check if the column header is in the GeoDataFrame
    if column_header not in df.columns:
        raise ValueError(f"The column header '{column_header}' does not exist in the GeoDataFrame.")
    
    # Check if the row index is within the bounds of the GeoDataFrame
    if not (0 <= int(row_index) < len(df)):
        raise ValueError(f"The row index {row_index} is out of bounds for the GeoDataFrame.")
    
    # Retrieve and return the value from the specified cell
    return df.iloc[row_index][column_header]
def count_rows_columns(df):
    """
    Counts the number of rows and columns in a pandas DataFrame.

    Parameters:
    - df (DataFrame): The pandas DataFrame whose dimensions will be counted.

    Returns:
    - tuple: A tuple containing two elements, the number of rows and the number of columns in the DataFrame.
    """
    rows, columns = df.shape  # df.shape returns a tuple (number of rows, number of columns)
    return rows, columns
def count_rows_in_excel(data_source, sheet_name=None):
    """
    Count rows in a given Excel sheet or pandas DataFrame.

    Parameters:
    - data_source (str or pd.DataFrame): The file path of the Excel file or a pandas DataFrame.
    - sheet_name (str, optional): The name of the sheet to count rows in. Used only if data_source is a file path.

    Returns:
    - int: Number of rows in the Excel sheet or DataFrame.
    """
    if isinstance(data_source, pd.DataFrame):
        # If data_source is a DataFrame, simply return the number of rows
        return len(data_source)
    elif isinstance(data_source, str):
        # Assume data_source is a file path to an Excel file
        try:
            wb = load_workbook(filename=data_source, read_only=True)
            if sheet_name and sheet_name in wb.sheetnames:
                ws = wb[sheet_name]
            else:
                ws = wb[wb.sheetnames[0]]

            row_count = ws.max_row
            if row_count == 1 and ws.max_column == 1:
                if ws.cell(row=1, column=1).value is None:
                    row_count = 0
            
            wb.close()
            return row_count
        except utils.exceptions.InvalidFileException:
            print("Failed to open the file. It may be corrupted or the path is incorrect.")
            return None
        except Exception as e:
            print(f"An error occurred: {e}")
            return None
    else:
        print("Invalid input. Please provide a valid file path or pandas DataFrame.")
        return None
def read_excel_range(df, start_row, num_rows):
    # Skip rows up to start_row, not including the header if it's the first row.
    # Adjust start_row by 1 if your Excel file has headers and you want to include them.
    # start_row is 0-indexed in Python, but 1-indexed in Excel, so adjust accordingly.
    skip = start_row - 1 if start_row > 0 else None
    
    # Read the specified range of rows
    df = get_df(df, skiprows=skip, nrows=num_rows, header=None if skip is None else 0)
    
    return df
def get_row(target_value,column_name=None,index_value=None,df=None):
    # Use the `.str.contains` method directly in the indexing to filter rows
    df = get_df(df)
    column_names = make_list(column_name or df.columns.tolist())
    for i,column in enumerate(column_names):
        search_column(target_value,column_name=None,column_index=0,df=None,exact_match=True)
        value = df[column][index_value]
        if value.lower() == target_value.lower():
            return df,i
    return None,None
def search_column(target_value,column_name=None,column_index=0,df=None,exact_match=True):
    df = get_df(df)
    headers = df.columns.tolist()
    if not column in headers and column_index >len(headers):
        return
    column_values = df[column].tolist()
    if target_value in column_values:
        return get_itter(column_values=column_values,target_value=target_value,exact_match=exact_match)
    if not exact_match:
        if [val for val in column_values if str(target_value) in str(val)]:
           return get_itter(column_values=column_values,target_value=target_value,exact_match=exact_match)
    column_names = make_list(column_name or headers)
    return column_names
def convert_value(value, column_dtype):
    """ Convert the value to the column data type if possible, handling string representations of numbers. """
    if pd.api.types.is_numeric_dtype(column_dtype):
        try:
            return float(value)
        except ValueError:
            return value  # Return the original value if conversion fails
    return value
















from .query_tools import *
def add_data_point(df, data_point,header=None):
    """
    Adds a data point and its header to the next available column in the Excel file.

    Parameters:
        file_path (str): The path to the Excel file.
        header (str): The header for the data point.
        data_point (str or numeric): The data point to add.
    
    Returns:
        None: The function updates the Excel file directly.
    """
    # Ensure the file exists or create a new DataFrame if it doesn't
    df = get_df(df)
    # Check if the header already exists
    if header in df.columns:
        # Find first empty row in the existing column
        empty_row = df[header].isna().idxmax() if not df[header].dropna().empty else 0
        df.at[empty_row, header] = data_point
    else:
        # Add new column with the header and place the data point at the first row
        df[header] = pd.Series([data_point] + [None] * (len(df) - 1))
    
    # Save the updated DataFrame back to the Excel file
    return df
def add_or_update_headers(df, column_name, default_value=None):
    """
    Add a new column to a DataFrame with a default value if it does not already exist.

    Parameters:
    df (DataFrame): The DataFrame to modify.
    column_name (str): The name of the column to add.
    default_value (Any, optional): The default value to assign to the new column. Defaults to None.

    Returns:
    DataFrame: The modified DataFrame with the new column added if it didn't exist.
    """
    if column_name not in df.columns:
        df[column_name] = default_value
    else:
        print(f"Column '{column_name}' already exists in the DataFrame. No changes made.")

    return df
def does_not_equal_in_list(obj,list_objs):
    for list_obj in list_objs:
        if obj == list_obj:
            return False
    return True
def update_or_append_data(df_new_data=None, df_existing_data=None, search_column=None, search_value=None, clear_duplicates=False):
    df_new = get_df(df_new_data)
    df_existing = get_df(df_existing_data)
    
    if df_existing.empty:
        return df_new

    # Ensure new data columns exist in the existing dataframe, add if not
    for col in df_new.columns:
        if col not in df_existing.columns:
            df_existing[col] = pd.NA  # Use pd.NA for missing data

    if search_column and search_value:
        if isinstance(search_column, list) and isinstance(search_value, list):
            mask = pd.Series(True, index=df_existing.index)
            for col, val in zip(search_column, search_value):
                mask &= (df_existing[col] == val)
        else:
            mask = (df_existing[search_column] == search_value)

        if does_not_equal_in_list(mask,['',' ',[],None,'none','None']):
            # Update existing rows based on mask
            for col in df_new.columns:
                df_existing.loc[mask, col] = df_new.loc[df_new.index[0], col]
            print(f"Updated rows where {search_column} matches {search_value}.")
        else:
            # Append new data if no matching row is found
            df_existing = pd.concat([df_existing, df_new], ignore_index=True)
            print(f"Appended new data as no existing match found for {search_value}.")

        # Handle duplicates update
        if not clear_duplicates:
            first_indices = df_existing.drop_duplicates(subset=search_column, keep='first').index
            update_mask = ~df_existing.index.isin(first_indices)
            for col in df_existing.columns:
                df_existing.loc[update_mask, col] = df_existing.loc[df_existing[df_existing[search_column] == df_existing.loc[update_mask, search_column]].index[0], col]

            print("Updated duplicates to match the first occurrence.")
    else:
        df_existing = pd.concat([df_existing, df_new], ignore_index=True)
        print("Appended new data as no search criteria provided.")

    if clear_duplicates:
        df_existing = df_existing.drop_duplicates()
        print("Duplicates removed after update/append.")

    return df_existing
def get_min_max_from_query(query):
    query = [return_float_or_int(obj) for obj in make_list(query) if is_number(obj)]
    query.sort()
    minimum = get_number(safe_get(query,0))
    maximum = get_number(safe_get(query,-1))
    minimum = minimum if len(query) > 0 and minimum is not None else 0
    maximum = maximum if len(query) > 1 and maximum is not None else 900
    return minimum,maximum

def search_df_for_values(df, column_name, query_list, type_dependent=False):
    """
    Search DataFrame column for rows matching any items in query_list with optional type-dependent matching.

    Parameters:
    - df (pd.DataFrame): The DataFrame to search.
    - column_name (str): The name of the column to search.
    - query_list (list or single value): A list of values or a single value to search for in the column.
    - type_dependent (bool): Whether to enforce type matching.

    Returns:
    - pd.DataFrame: A DataFrame of rows where the column values match any item in the query_list.
    """
    df = pd.DataFrame(df)  # Ensure it is a DataFrame
    query_list = make_list(query_list)

    if type_dependent:
        # Enforcing exact type matching
        mask = df[column_name].apply(lambda x: any([x == item and type(x) == type(item) for item in query_list]))
    else:
        # Attempt to convert query values to the column data type for accurate comparison
        column_dtype = df[column_name].dtype
        converted_query_list = [convert_value(item, column_dtype) for item in query_list]
        mask = df[column_name].isin(converted_query_list)

    return df[mask]

def search_df_with_condition(df, column_name, condition_func):
    """
    Search DataFrame column to find rows where condition_func returns True.

    Parameters:
    - df (pd.DataFrame): The DataFrame to search.
    - column_name (str): The column to apply the condition on.
    - condition_func (function): A function that takes a single value and returns True or False.

    Returns:
    - pd.DataFrame: A DataFrame of rows where the column values satisfy the condition_func.
    """
    df=get_df(df)
    # Applying the condition function vectorized
    mask = df[column_name].apply(condition_func)
    return df[mask]
def query_dataframe(df, query_string):
    """
    Use DataFrame.query() to filter rows based on a query string.

    Parameters:
    - df (pd.DataFrame): The DataFrame to query.
    - query_string (str): The query string to evaluate.

    Returns:
    - pd.DataFrame: The filtered DataFrame.
    """
    return df.query(query_string)
def filter_and_deduplicate_df(df, filter_columns, filter_values, dedup_columns=None):
    """
    Filters a DataFrame based on specified values in given columns and removes duplicates.

    Parameters:
    - df (pd.DataFrame): The DataFrame to filter and deduplicate.
    - filter_columns (list of str): Column names to apply the filters on.
    - filter_values (list of list): Lists of values to include for each column in filter_columns.
    - dedup_columns (list of str, optional): Columns to consider for dropping duplicates. If not specified,
      duplicates will be dropped based on all columns.

    Returns:
    - pd.DataFrame: The filtered and deduplicated DataFrame.
    """
    # Ensure the input integrity
    assert len(filter_columns) == len(filter_values), "Each filter column must correspond to a list of filter values."

    # Apply filters based on the columns and corresponding values
    mask = pd.Series([True] * len(df))
    for col, vals in zip(filter_columns, filter_values):
        mask &= df[col].isin(vals)

    filtered_df = df[mask]

    # Drop duplicates based on specified columns
    if dedup_columns:
        deduplicated_df = filtered_df.drop_duplicates(subset=dedup_columns)
    else:
        deduplicated_df = filtered_df.drop_duplicates()

    return deduplicated_df
async def get_range(df, column,query,inverseOption=False,caseOption=False,substringOption=False):
    """
    Filters the DataFrame based on numeric ranges specified in query.
    Optionally inverts the filter to exclude the specified range.
    
    :param df: DataFrame or path to DataFrame
    :param column: Column name to apply the range filter on
    :param query: List or tuple containing the minimum and maximum values as strings
    :param invert: If True, the range is inverted (excludes the range specified)
    :return: Filtered DataFrame
    """
    if not isinstance(query,list):
        query = split_and_clean_lines(query)
    logging.info(f"get_range processing")
    temp_df = df.copy()
    logging.info(f"temp_df created with type {type(temp_df)}")
    minimum,maximum = get_min_max_from_query(query)
    logging.info(f"minimum,maximum : {minimum},{maximum}")
    temp_df[column] = pd.to_numeric(temp_df[column], errors='coerce')  # Coerce errors in case of non-numeric data
    # Create a condition for values within the specified range
    condition = temp_df[column].between(minimum,maximum , inclusive='both')
    # Filter the original DataFrame using the condition from the temporary DataFrame
    series = is_inverse(condition,inverse=inverseOption)
    return df[series]
async def filter_dataframe(df, column,query,caseOption=False,substringOption=False,inverseOption=False,splitQuery=False):
    """
    Filters a DataFrame based on a comparison between a column's values and a list.

    Parameters:
    - df: Pandas DataFrame.
    - column: Column name as a string where the comparisons are to be made.
    - compare_list: List of strings to compare against the DataFrame's column.
    - substring_match: Boolean, if True performs substring matching, otherwise exact matching.
    - case_sensitive: Boolean, if False converts both column and list_obj to lowercase.

    Returns:
    - filtered_df: DataFrame containing only the rows that meet the condition.
    """
    if not isinstance(query,list):
        if splitQuery:
            
            if not isinstance(splitQuery,bool) and str(splitQuery) in str(query):
                logging.info(f"splitQuery chosen; splitting query with general {splitQuery}...")                
                query = clean_list(query.split(splitQuery))
            else:
                logging.info(f"splitQuery chosen; splitting query with general split...")
                query = split_and_clean_lines(query)
    compare_list=make_list(query)
    
    if not caseOption:
        logging.info(f"caseOption not chosen; normalizing case...")
        compare_list = [str(item).lower() for item in compare_list]
        df[column] = df[column].astype(str).str.lower()

    if substringOption:
        logging.info(f"substringOption chosen; searching for values in query object")
        condition = df[column].apply(lambda x: any(str(item) in str(x) for item in compare_list))
    else:
        logging.info(f"substringOption not chosen; searching for values == to query object")
        condition = df[column].apply(lambda x: any(str(item) == str(x) for item in compare_list))
    series = is_inverse(condition,inverse=inverseOption)
    return df[series]

async def filter_excel(df,column,query,range_option=False,caseOption=False,substringOption=False,inverseOption=False,distanceOption=False,geo_location_refference=None):
    logging.info(f"recieved excel: parsing for column={column},range_option={range_option},caseOption={caseOption},substringOption={substringOption},inverseOption={inverseOption},distanceOption={distanceOption},geo_location_refference={geo_location_refference}")
    df = get_df(df)
    logging.info(f"query came as {query}")
    if not isinstance(query,list):
        query = split_and_clean_lines(query)
    query=make_list(query)
    logging.info(f"query started as  as {query}")
    if distanceOption:
        logging.info(f"distanceOption selected")
        column=column or 'ZIP'
        query = await distance_within_range(distance=query,geo_location_refference=geo_location_refference)
        logging.info(f"query is now  as {query}")
    if range_option:
        logging.info(f"rangeOption selected")
        df = await get_range(df,column,query,inverseOption=inverseOption,caseOption=caseOption,substringOption=substringOption)
    else:
        df = await filter_dataframe(df,column,query,caseOption=caseOption,substringOption=substringOption,inverseOption=inverseOption)
    return df



def get_geo_location_master():
    return "/home/catalystRepository/distance_repository/filters/geo_location/zip_codes/sacramento_to_zipcodes_geo_locations.xlsx"
def get_geo_location_master_headers():
    return """COUNTY	COUNTY_closest_latitude	COUNTY_closest_longitude	ZIP	ZIP_closest_latitude	ZIP_closest_longitude	ZIP_COUNTY_closest_distance(mi)	ZIP_furthest_latitude	ZIP_furthest_longitude	ZIP_COUNTY_furthest_distance(mi)""".split('\t')
async def distance_within_range(distance,geo_location_refference=None):
    logging.info(f"distance = {distance}\n\geo_location_refference={geo_location_refference}")
    geo_location_refference = geo_location_refference or get_geo_location_master()
    # Ensure distance is a float
    distance = float(get_number(safe_get(distance, 0)))
    
    # Load geo location data
    df_geo_location = get_df(geo_location_refference)
    
    # Convert ZIP codes to integers, assuming all entries are valid zip codes
    df_geo_location['ZIP'] = df_geo_location['ZIP'].astype(str)
    
    # Convert distances to floats
    df_geo_location['ZIP_COUNTY_furthest_distance(mi)'] = df_geo_location['ZIP_COUNTY_furthest_distance(mi)'].astype(float)
    df_geo_location['ZIP_COUNTY_closest_distance(mi)'] = df_geo_location['ZIP_COUNTY_closest_distance(mi)'].astype(float)
    
    # Create a mask where the distance is within the specified range
    mask = (distance >= df_geo_location['ZIP_COUNTY_closest_distance(mi)']) | (distance >= df_geo_location['ZIP_COUNTY_furthest_distance(mi)'])
    logging.info(f"mask = {mask}")

    # Filter zip codes based on the mask
    good_zips = df_geo_location.loc[mask, 'ZIP'].tolist()
    good_zips=['00000'[:-len(zips)]+str(zips) for zips in good_zips]
    return good_zips



import numpy as np
import cv2

# Read input image
img = cv2.imread('/home/joben/Desktop/solar/cropped_thermal_areas/blackened.jpg')
# Find contours in thresh (find the triangles).
contours = cv2.findContours(thresh, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)[-2]  # [-2] indexing takes return value before last (due to OpenCV compatibility issues).

# Iterate triangle contours
for c in contours:
    if cv2.contourArea(c) > 4:  #  Ignore very small contours
        # Mark triangle with blue line
        cv2.drawContours(img, [c], -1, (255, 0, 0), 2)
# Convert from BGR to HSV color space
hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)

# Get the saturation plane - all black/white/gray pixels are zero, and colored pixels are above zero.
s = hsv[:, :, 1]

# Apply threshold on s - use automatic threshold algorithm (use THRESH_OTSU).
ret, thresh = cv2.threshold(s, 0, 255, cv2.THRESH_BINARY+cv2.THRESH_OTSU)

# Find contours in thresh (find only the outer contour - only the rectangle).
contours = cv2.findContours(thresh, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)[-2]  # [-2] indexing takes return value before last (due to OpenCV compatibility issues).

# Mark rectangle with green line
cv2.drawContours(img, contours, -1, (0, 255, 0), 2)

# Assume there is only one contour, get the bounding rectangle of the contour.
x, y, w, h = cv2.boundingRect(contours[0])

# Invert polarity of the pixels inside the rectangle (on thresh image).
thresh[y:y+h, x:x+w] = 255 - thresh[y:y+h, x:x+w]

# Find contours in thresh (find the triangles).
contours = cv2.findContours(thresh, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)[-2]  # [-2] indexing takes return value before last (due to OpenCV compatibility issues).

# Iterate triangle contours
for c in contours:
    if cv2.contourArea(c) > 4:  #  Ignore very small contours
        # Mark triangle with blue line
        cv2.drawContours(img, [c], -1, (255, 0, 0), 2)

# Show result (for testing).
cv2.imshow('img', img)
cv2.waitKey(0)
cv2.destroyAllWindows()
image_path = '/home/joben/Desktop/solar/cropped_thermal_areas/blackened.jpg'
outline_black_shapes(image_path)




#!/usr/bin/env python3
"""
extract_table_from_pdf.py

Illustration of how to take character‐level PDFMiner output and group it
into rows and columns, then save as CSV or Excel.

Usage:
    python3 extract_table_from_pdf.py input.pdf output.xlsx

Dependencies: pdfminer.six, pandas, openpyxl
    pip install pdfminer.six pandas openpyxl
"""

import sys
import os
from typing import List, Dict, Any, Tuple
import pandas as pd

from pdfminer.pdfparser import PDFParser
from pdfminer.pdfdocument import PDFDocument
from pdfminer.pdfpage import PDFPage
from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
from pdfminer.converter import PDFPageAggregator
from pdfminer.layout import LAParams, LTChar, LTAnno

# -----------------------------------------------------------------------------------
def extract_chars(pdf_path: str) -> List[Dict[str, Any]]:
    """
    Return a list of dicts, one per LTChar, containing page, char, x0, y0, x1, y1.
    """
    records: List[Dict[str, Any]] = []

    with open(pdf_path, "rb") as fp:
        parser = PDFParser(fp)
        doc = PDFDocument(parser)
        rsrcmgr = PDFResourceManager()
        laparams = LAParams()
        device = PDFPageAggregator(rsrcmgr, laparams=laparams)
        interpreter = PDFPageInterpreter(rsrcmgr, device)

        for page_number, page in enumerate(PDFPage.create_pages(doc), start=1):
            interpreter.process_page(page)
            layout = device.get_result()

            def recurse(obj):
                if isinstance(obj, LTChar):
                    # record every non‐whitespace character
                    txt = obj.get_text()
                    if not txt.isspace():  # skip pure spaces/newlines
                        records.append({
                            "page": page_number,
                            "char": txt,
                            "x0": round(obj.x0, 3),
                            "y0": round(obj.y0, 3),
                            "x1": round(obj.x1, 3),
                            "y1": round(obj.y1, 3),
                        })
                else:
                    try:
                        for child in obj:
                            recurse(child)
                    except Exception:
                        pass

            for element in layout:
                recurse(element)

    return records


# -----------------------------------------------------------------------------------
def cluster_rows(chars: List[Dict[str, Any]], y_tolerance: float = 2.0) -> List[List[int]]:
    """
    Group all character‐indices (in the `chars` list) into “rows” based on y1.
    Returns a list of clusters; each cluster is a list of integer indices into `chars`.

    y_tolerance: any two characters whose y1 differ by <= this amount → same row.
    """
    # First, collect (index, y1) pairs and sort by y1 DESC (top of page first)
    indexed = [(i, c["y1"]) for i, c in enumerate(chars)]
    indexed.sort(key=lambda x: x[1], reverse=True)

    row_clusters: List[List[int]] = []
    for idx, y1 in indexed:
        placed = False
        # Try to put this character into an existing row cluster
        for cluster in row_clusters:
            # `cluster` is a list of indices; check the y1 of the first char in that cluster
            first_idx = cluster[0]
            if abs(chars[first_idx]["y1"] - y1) <= y_tolerance:
                cluster.append(idx)
                placed = True
                break
        if not placed:
            row_clusters.append([idx])

    # Now each cluster is a “row” of character‐indices; we can return them
    return row_clusters


# -----------------------------------------------------------------------------------
def cluster_columns_for_row(
    chars: List[Dict[str, Any]],
    row_indices: List[int],
    x_tolerance: float = 2.0
) -> List[List[int]]:
    """
    Given a list of indices (all on the same row), group them into columns.
    We sort by x0 and say: “If the next character’s x0 is more than x_tolerance away
    from the previous cluster’s right‐edge, start a new column.”

    Returns: a list of clusters (one cluster = list of indices belonging to that column).
    """
    # Build a list of (idx, x0, x1), sorted by x0 ascending
    slice_info: List[Tuple[int, float, float]] = []
    for i in row_indices:
        slice_info.append((i, chars[i]["x0"], chars[i]["x1"]))
    slice_info.sort(key=lambda x: x[1])  # sort by x0

    col_clusters: List[List[int]] = []
    for idx, x0, x1 in slice_info:
        if not col_clusters:
            col_clusters.append([idx])
        else:
            # Look at the rightmost x1 of the last cluster
            last_cluster = col_clusters[-1]
            rightmost = max(chars[j]["x1"] for j in last_cluster)
            if x0 - rightmost <= x_tolerance:
                # This char overlaps (or is very close) with previous cluster horizontally
                last_cluster.append(idx)
            else:
                # Far enough that we treat it as a new column
                col_clusters.append([idx])

    return col_clusters


# -----------------------------------------------------------------------------------
def rows_and_columns_from_chars(
    chars: List[Dict[str, Any]]
) -> List[List[str]]:
    """
    Given all character records on a single page, return a 2D list:
        [ [cell1_row1, cell2_row1, ...],
          [cell1_row2, cell2_row2, ...],
          ... ]

    Steps:
      1) cluster_rows → list of “rows” (each row is a list of character‐indices)
      2) for each row, cluster_columns_for_row → list of “columns” (each a list of indices)
      3) within each column, sort indices by x0, concatenate chars in that order
    """
    table_rows: List[List[str]] = []

    # Cluster characters into rows (by y1)
    row_clusters = cluster_rows(chars)

    for row_idx_list in row_clusters:
        # For each row, cluster into columns
        col_clusters = cluster_columns_for_row(chars, row_idx_list)

        # Now build a list-of-strings for each column in that row
        row_cells: List[str] = []
        for col in col_clusters:
            # Sort all char‐indices in this column by x0 so they read left→right
            sorted_col = sorted(col, key=lambda i: chars[i]["x0"])
            cell_text = "".join(chars[i]["char"] for i in sorted_col)
            row_cells.append(cell_text.strip())

        table_rows.append(row_cells)

    return table_rows


# -----------------------------------------------------------------------------------
def convert_pdf_to_table_dataframe(pdf_path: str) -> pd.DataFrame:
    """
    Extract all chars, group into rows/columns, then return a DataFrame.
    Each row of the DataFrame is one “table row,” with columns named “col_1, col_2, ...”.
    """
    chars = extract_chars(pdf_path)
    if not chars:
        return pd.DataFrame()

    # Process page by page. Here we do everything in one DF with a 'page' column.
    # You could also split by page if needed.
    df_list: List[pd.DataFrame] = []
    for page_number in sorted({c["page"] for c in chars}):
        page_chars = [c for c in chars if c["page"] == page_number]

        # Group into rows & columns
        table_rows = rows_and_columns_from_chars(page_chars)

        # Build a DataFrame; pad shorter rows so the table is rectangular
        max_cols = max(len(r) for r in table_rows) if table_rows else 0
        normalized = [row + [""]*(max_cols - len(row)) for row in table_rows]

        page_df = pd.DataFrame(normalized, columns=[f"col_{i+1}" for i in range(max_cols)])
        page_df.insert(0, "page", page_number)
        df_list.append(page_df)

    # Concatenate all pages’ tables
    result_df = pd.concat(df_list, ignore_index=True)
    return result_df


# -----------------------------------------------------------------------------------
def pdf_to_excel(input_pdf,output_xlsx=None):
    input_pdf = file
    ext = input_pdf.split('.')[-1]
    output_xlsx = output_xlsx or f"{input_pdf[:-len(ext)]}.xlsx"
    if not os.path.isfile(input_pdf):
        print(f"ERROR: PDF not found: {input_pdf}")
        sys.exit(1)
    table_df = convert_pdf_to_table_dataframe(input_pdf)
    return table_df


